package ec.app.sequence;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Random;
import java.util.StringTokenizer;

import org.apache.commons.io.FileUtils;
/**
 * This is generic Utility to parse RNA sequences, with exons data given, to find the exact locations
 * of Acceptors and Donors. This is useful for generating 'true sequences' when exons are known.
 * The Data file needs to be in format of sequenceId \t Sequence. Exons data file need to be in
 * format of sequenceId start1 end1 \n, sequenceId start2 end2 etc
 * @author udaykamath
 *
 */
public class ExonParser {
	public static HashMap<String,List<int[]>> organismToExonPositions = new HashMap<String,List<int[]>>();
	public static String AG_SEQUENCE= "ag";
	public static String GT_SEQUENCE="gt";
	protected static final char[] ACTG = new char[] { 'a', 't', 'g', 'c' };
	protected static final Random random = new Random();
	protected static List<String> donorList = new ArrayList<String>();
	protected static List<String> acceptorList = new ArrayList<String>();
	
	public static void addToMap(String organismId, String start, String end){
		List<int[]> currentList = organismToExonPositions.get(organismId);
		if(currentList == null){
			//add a list to this organism
			currentList = new ArrayList<int[]>();
		}
		//add the positions
		currentList.add(new int[]{new Integer(start).intValue()-1, new Integer(end).intValue()-1});
		//put it back
		organismToExonPositions.put(organismId,currentList);
	}
	
	public static String getFillers(int size){
		String filler = "";
		char[] fillerCharacters = new char[size];
		for(int i = 0; i< size; i++){
			fillerCharacters[i] = ACTG[random.nextInt(3)];
		}
		filler = new String(fillerCharacters);
		return filler;
	}
	
	private static String getDonorSplicedSequence(int position, String sequence){
		String totalSequence = null;
		if(position > 0 && position +2 < sequence.length()){
			String shortDonorSequence = sequence.substring(position+1, position+3);
			
			//only if its donor
			if(shortDonorSequence.equalsIgnoreCase(GT_SEQUENCE)){
				int countUpStream = position + 3 +80;
				String fillerUpStream = null;
				if(countUpStream > sequence.length()){
					//need fillers 
					fillerUpStream = getFillers(Math.abs(countUpStream- sequence.length()));
				}
				String upStreamSequence ="";
				if(fillerUpStream != null){
					String partOfSubStreamSequence = sequence.substring(position+ 3, sequence.length());
					upStreamSequence = partOfSubStreamSequence + fillerUpStream;
				}
				else 
					upStreamSequence = sequence.substring(position+3, position+3 +80);
				int countDownStream = position+1 - 80;
				String fillerDownStream = null;
				if(countDownStream < 0){
					//need fillers 
					fillerDownStream = getFillers(Math.abs(countDownStream));
				}
				String downStreamSequence ="";
				if(fillerDownStream != null){
					String partOfSubStreamSequence = sequence.substring(0, position +1);
					downStreamSequence = fillerDownStream + partOfSubStreamSequence;
				}
				else
					downStreamSequence = sequence.substring(position-79, position+1);
			
				totalSequence =downStreamSequence + GT_SEQUENCE + upStreamSequence;
			}
		}
		return totalSequence;
	}
	
	private static String getAcceptorSplicedSequence(int position, String sequence){
		String totalSequence = null;
		if(position > 0){
			String shortDonorSequence = sequence.substring(position-2, position);			
			//only if its acceptor or donor
			if(shortDonorSequence.equalsIgnoreCase(AG_SEQUENCE)){
				int countUpStream = position  +80;
				String fillerUpStream = null;
				if(countUpStream > sequence.length()){
					//need fillers 
					fillerUpStream = getFillers(Math.abs(countUpStream- sequence.length()));
				}
				String upStreamSequence ="";
				if(fillerUpStream != null){
					String partOfSubStreamSequence = sequence.substring(position, sequence.length());
					upStreamSequence = partOfSubStreamSequence + fillerUpStream;
				}
				else 
					upStreamSequence = sequence.substring(position, position +80);
				int countDownStream = position -2 -80;
				String fillerDownStream = null;
				if(countDownStream < 0){
					//need fillers 
					fillerDownStream = getFillers(Math.abs(countDownStream));
				}
				String downStreamSequence ="";
				if(fillerDownStream != null){
					String partOfSubStreamSequence = sequence.substring(0, position-2);
					downStreamSequence = fillerDownStream + partOfSubStreamSequence;
				}
				else
					downStreamSequence = sequence.substring(position-2-80, position-2);
				
				totalSequence =downStreamSequence + AG_SEQUENCE + upStreamSequence;
			}
		}
		return totalSequence;
	}
	
	public static void parseSequence(String chromosomeId, String sequence){
		//now search for the id in our exon position
		List<int[]> positions = organismToExonPositions.get(chromosomeId);
		//now lets get through all positions
		for(int i =0; i< positions.size(); i++){
			//get the position i
			int[] exonStartAndStop = positions.get(i);
			int start = exonStartAndStop[0];
			int stop = exonStartAndStop[1];
			//inspect stop for trigger
			String donorSequence = getDonorSplicedSequence(stop, sequence);
			if(donorSequence != null){
				System.out.println("DONOR" + " " +stop + " " + donorSequence);
				donorList.add(donorSequence +" 1 " + chromosomeId + " "  +stop);
			}
			String acceptorSequence = getAcceptorSplicedSequence(start, sequence);
			if(acceptorSequence != null) {
				System.out.println("ACCEPTOR" + " " +start + " " + acceptorSequence);
				acceptorList.add(acceptorSequence +" 1 " + chromosomeId + " "  +start);
			}
			
			
			
			
		}
		
	}
	
	public static void main(String[] args) throws Exception{
		if(args.length <2){
			System.err.println("ExonParser exonDataFile sequenceFile");
			System.exit(-1);
		}
			
		//get the exon data file
		File exonDataFile = new File(args[0]);
		//read the descriptionfile
		List dataDescription = FileUtil.readTextFileToStringList(exonDataFile, false);
		for(int i =0; i <dataDescription.size(); i++){
			//get the data
			String dataLine = (String)dataDescription.get(i);
			//if its not empty
			if((dataLine.length() >0)){
				//tokenize it
				StringTokenizer stk = new StringTokenizer(dataLine);
				//get the id
				String organismId = stk.nextToken();
				//get the start 
				String startPosition = stk.nextToken();
				//get the end
				String endPosition = stk.nextToken();
				//pass it on 
				addToMap(organismId, startPosition, endPosition);
			}
		}
		//read the sequences
		BufferedReader reader = new BufferedReader(new InputStreamReader(
				new FileInputStream(args[1])));
		//read the first line
		String totalLine = reader.readLine();
		while (totalLine != null) {
			int startOfSpace = totalLine.indexOf(" ");
			String chromosomeId = totalLine.substring(0, startOfSpace);
			String sequence = totalLine.substring(startOfSpace + 1, totalLine
					.length());
			//now let the parser understand what to do
			parseSequence(chromosomeId, sequence);
			totalLine = reader.readLine();
		}
		//print the donorsamples[2]
		FileUtils.writeLines(new File(args[2]), donorList);
		FileUtils.writeLines(new File(args[3]), acceptorList);
	}

}
